/*  $Id: fconfig.cc,v 1.17 1999/12/18 21:32:36 shaggy Exp $
 *      file configuration utility
 *      
 *      relative path: ./fconfig
 *  
 *  Copyright (c) 1995-1998 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#define fconfig_VER_MAJOR	0
#define fconfig_VER_MINOR	34
#define fconfig_VER_PATCH	0
#define fconfig_VER_STR	"0.34"
#define fconfig_VER_DATE	"June 30, 1999"
#define fconfig_VER_NDATE	"06/30/99"


#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <dirent.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>

typedef char sz;
typedef const char csz;

// it seems that some systems (like old Ultrix) don't have these defined,
// even though they're in the libraries
extern "C" sz* strdup(csz *);
extern "C" int symlink(csz *, csz *);


#ifndef PATH_MAX
#define PATH_MAX	1024	// pathname max size
#endif
#ifndef NAME_MAX
#define NAME_MAX	256	// filename max size
#endif
#define VNAME_MAX	128	// variable name max size
#ifndef LINE_MAX
#define LINE_MAX	256	// input line max size
#endif
#define CMD_MAX		8192	// command buffer max size
#define BUF_MAX		65536	// temporary buffer max size
#define MAX_FLAGS	128	// max # of flags


// notes:
//	allow files names "file.C.head" that will
//	be included prior to all "file.C" files
//	give option to join files physically together
//	or use cpp to join them at compile time

// called with:
//	fconfig (machtypes...) (src...) files ...
//		machtypes: a single argument that is a list of
//			   of machine type specifications
//		src:	   list of paths from the current
//			   directory to each source directory
//		files:	   all remaining arguments are the names
//			   of the source files to configure
// uses environment variables:
//	Autoconfig: set this flag for auto-configuration (0 or false disables)
//	Command_c: (gcc) command for compiling .c files	(gcc)
//	Command_C: (gxx) command for compiling .C files	(g++)
//	Command_S: (as)	command for compiling .S files 	(as)
//	etc...
//	Flags_Needed:	flags needed for any file type
//	Flags_Needed_c:	defaults flags needed for .c files
//	Flags_Wanted_c: desired flags for .c files
//	Flags_Banned_c: flags banned from .c files
//	Flags_Needed_gxx:	flags needed for g++
//	Makedepend_c:	command for making .c dependencies (gcc -M)
//	Cpp_c:		command for running cpp on .c files (gcc -E)
//
// variable substitutions:
//	$@	target name
//	$<	source name
//	$^	names of all dependencies? (not supported)
//	$*	the stem of the filename
//
// job:
//	searches source area and all appropriate subdirectories
//	for files with the same name
//	i.e.:	machtype = i486 pc i386
//	search includes: . m-i386 m-i486 m-pc m-vax-i386 m-68k-i486 etc
//	builds a local "file.C" that contains #include references
//	to all the other parts
//	run through "gcc -M" to get dependency listing
//	run through "gcc -E | grep "#pragma Make"" to get all commands
//	define env vars for each pragma that starts with Make
//	check for conflicts, and generate command and flags
//	for building each target
//
//	Generate "file.o.mak" that consists of:
//		file.o file.o.mak: file.C ../../src/blah/file.C etc...
//		file.o: 
//			gcc -c -O2 -fforce-mem -funroll-loops file.C
//		file.o.mak:
//			fconfig "m-i386 m-pc" "../../src" blah file.C
//
//	pragma directives:
//		#pragma Make Local  (or any other string without an '=')
//		#pragma	Make Command = gcc -c
//		#pragma	Make Flags_Needed = -fforce-mem
//		#pragma Make Flags_Wanted = -funroll-loops
//		#pragma	Make Flags_Banned = -pedantic
//	future (maybe):
//		#pragma Main #pragma implementation
//		#pragma Implementation
//		#pragma Interface
// errors:
//	o	should only be one "Command_ext_etc.." var for each
//			file extension (ext)
//	o	should never have conflict of flags needed and
//			flags banned


csz *const *machtypes;
int nmach;
csz *const *src;
int nsrc;
csz *const *argvp;

csz *target_name;	// target filename
csz *source_name;	// source filename
csz *base_name;		// current basename (stem) of source
sz *extensions;		// string of possible extensions
bool dest_makefile;		// true if writing flags to makefile

FILE *logfile;

bool autoconfig = false;
bool warnings = true;


class DirQ {
	DirQ	*_next;
	DirQ	*_prev;
	csz	*_path;
	csz	*_name;
	csz	*_ref;	// reference extension (i.e. m-i386-d2)
	csz	*_ext;	// source extension expected
public:
	DirQ()
		{ _name = 0; _next = this; _prev = this; }
	DirQ(csz *newpath, csz *newname, csz *newref)
		{ _path = newpath; _name = newname; _ref = newref;
		  _next = this; _prev = this; _ext = ""; }
	DirQ(csz *newpath, csz *newname, csz *newref, csz *newext)
		{ _path = newpath; _name = newname; _ref = newref;
		  _next = this; _prev = this; _ext = newext; }
	void	add_head(DirQ *d)
		{ d->_next = _next; d->_prev = this; _next->_prev = d; _next = d; }
	void	add_tail(DirQ *d)
		{ d->_prev = _prev; d->_next = this; _prev->_next = d; _prev = d; }
	DirQ*	next() const
		{ return _next->_name ? _next : 0; }
	DirQ*	prev() const
		{ return _prev->_name ? _prev : 0; }
	DirQ*	remove() 
		{ _next->_prev = _prev; _prev->_next = _next;
		  _prev = 0; _next = 0; return this; }
	csz*	name()	const { return _name; }
	csz*	path() const { return _path; }
	csz*	ref() const	{ return _ref; }
	csz*	ext() const	{ return _ext; }
};


DirQ	dirs;


bool
file_exists(csz *fname)
{
	struct stat stbuf;

	return stat(fname, &stbuf) == 0;
}



csz*
getnext(csz **ap, sz *buf)
{
	sz *bp = buf;
	if(!*ap)
		return 0;
	while(**ap && **ap <= ' ')
		(*ap)++;
	if(!**ap)
		return 0; 
	while(**ap && **ap > ' ')
		*bp++ = *(*ap)++;
	*bp = '\0';
	return buf;
}

csz*
getnext_ext(csz **ap, csz *name, sz *buf)
{
	sz *bp = buf;
	while(*name)
		*bp++ = *name++;
	if(!*ap) 
		return 0;
	*bp = '_';
	if(!getnext(ap, bp+1)) {
		*ap = 0;	
		*bp = '\0';
	}
	return buf;
}


csz * const *
parse_list(csz *str, int *size)
{
	csz *cp;
	sz buf[VNAME_MAX];
	int count;
	sz **list;
	int i;

	cp = str;
	count = 0;
	while(getnext(&cp, buf)) 
		count++;
	list = new char*[count+1];
	cp = str;
	for(i=0; i<count; i++) 
		list[i] = strdup(getnext(&cp, buf));
	list[i] = 0;
	*size = count;
	return list;
}


bool
mach_match(csz *type, csz *name)
{
	csz *cp, *tp;
	if(strcmp(type, name) == 0)
		return true;
	if(name[0] != 'm' || name[1] != '-')
		return false;
	cp = name;
	while(*cp) {
		while(*cp != '-')
			if(!*cp++)
				return false;
		for(tp = type; *(++cp) == *tp && *tp; tp++)
			continue;
		if(!*tp && (!*cp || *cp == '-')) 
			return true;
	}
	return false;
}


void
find_dirs()
{
	int srcdir;
	DIR *dptr;
	struct dirent *d;
	DirQ dq;
	DirQ *dqp, *dnext;
	sz path[PATH_MAX];
	sz dname[NAME_MAX];
	sz ref[NAME_MAX];
	int mach;

	for(srcdir=nsrc-1; srcdir>=0; srcdir--) {
		int x;
		x = sprintf(path, "%s/", src[srcdir]);
		while(x > 0 && path[x-1] == '/')
			path[--x] = '\0';
		if(path[0] == '.') 
			strcpy(path, path+1+(path[1] == '/'));
		if(!(dptr = opendir(path[0] ? path : "."))) {
			perror(path);
			continue;
		}
		if(srcdir)
			sprintf(dname, "-d%d", srcdir);
		else
			dname[0] = '\0';
		while((d = readdir(dptr))) {
			sprintf(ref, "-%s%s", d->d_name, dname);
			if(!strncmp(d->d_name, "m-", 2))  
				dq.add_tail(new DirQ(path,
				    strdup(d->d_name), strdup(ref)));
		}
		closedir(dptr);
		sprintf(ref, "-mi%s", dname);
		dq.add_tail(new DirQ(path, ".", strdup(ref)));
	}
	for(mach=0; mach<nmach; mach++) {
		dqp = dq.next();
		if(strcmp(machtypes[mach], "gen") == 0)
			dirs.add_tail(new DirQ("./", ".", "-mg", ".mg"));
		while(dqp) {
			dnext = dqp->next();
			if(mach_match(machtypes[mach], dqp->name()))
				dirs.add_tail(dqp->remove());
			dqp = dnext;
		}
	}
}



void
convert_fname(sz *fname)
{
	while(*fname) {
		if((*fname < 'a' || *fname > 'z') 
				&& (*fname < 'A' || *fname > 'Z')
				&& (*fname < '0' || *fname > '9'))
			*fname = '_';
		fname++;
	}
}


sz* get_var(csz *name, sz *varbuf);


void
strcat_subst(sz *buf, csz *str)
{
	csz *src;
	sz varname[VNAME_MAX];
	sz *bp;

	while(*buf)
		buf++;
	while(*str)	
		if(*str == '$') {
			switch(str[1]) {
			case '@':	src = target_name; str++; break;
			case '<':	src = source_name; str++; break;
			case '*':	src = base_name; str++; break;
			case '(':
				str++;
				bp = varname; 
				while(*(++str) != ')')
					*bp++ = *str;
				*bp = '\0';
				get_var(varname, buf);
				src = buf;
				break;
			default:	
				if(dest_makefile)	
					src = "$$";
				else
					src = "$";
			}
			str++;
			while(*src)
				*buf++ = *src++;
		} else
			*buf++ = *str++;
	*buf = '\0';
}



const int WANTED	= 1;
const int NEEDED	= 2;
const int BANNED	= 4;

const int NOTUSED	= 0;
const int USED		= 1;
const int CONFLICT	= 2;


const int flag_action[] =
	{NOTUSED, USED, USED, USED,
	NOTUSED, NOTUSED, CONFLICT, CONFLICT};


static struct {
	sz	str[VNAME_MAX];
	int	type;
} flaglist[MAX_FLAGS];

static sz	conflictbuf[CMD_MAX];
static int	nflags;


void
add_flags(csz *fmt, csz *name, int type)
// search 'extensions' and match all combinations of
//	'name_ext1_ext2' and 'name_ext'
{
	csz *ext = extensions;
	csz *ext2;
	csz *varname;
	csz *varname2;
	sz buf[VNAME_MAX];
	sz buf2[VNAME_MAX];
	sz namestr[VNAME_MAX];
	csz *val;
	int i;
	csz *flag;
	sprintf(namestr, fmt, name);
	while((varname2 = getnext_ext(&ext, namestr, buf))) {
	    varname = varname2;
	    ext2 = extensions;
	    do {
	        if((val = getenv(varname))) {
		    while((flag = getnext(&val, buf))) {
			for(i=0; i<nflags; i++) {
				if(strcmp(flaglist[i].str, flag) == 0) {
					flaglist[i].type |= type;
					break;
				}
			}
			if(i == nflags) {
				strcpy(flaglist[i].str, flag);
				flaglist[i].type = type;
				nflags++;
			}
		    }

	        }
	    } while((varname = getnext_ext(&ext2, varname2, buf2)));
	}
}


sz*
get_flags(csz *name, sz *flagbuf)
{
	nflags = 0;
	add_flags("%s_Wanted", name, WANTED);
	add_flags("%s_Needed", name, NEEDED);
	add_flags("%s_Banned", name, BANNED);
	strcpy(flagbuf, "");
	strcpy(conflictbuf, "");
	for(int i=0; i<nflags; i++) 
		switch(flag_action[flaglist[i].type]) {
		case USED:
			strcat(flagbuf, " ");
			strcat_subst(flagbuf, flaglist[i].str);
			break;
		case CONFLICT:
			strcat(conflictbuf, " ");
			strcat(conflictbuf, flaglist[i].str);
			break;
		default:
			break;
		}
	if(conflictbuf[0])
		printf("Warning: Conflicting %s flags: %s\n",
			name, conflictbuf);
	return flagbuf;
}



sz*
get_var(csz *name, sz *varbuf)
// search 'extensions' and first try to match
//	'name_ext1_ext2'
//	and next just 'name_ext'
//	trying all possible extensions in order
//	if that fails, then it calls get_flags instead
{
	sz buf[VNAME_MAX];
	sz buf2[VNAME_MAX];
	csz *val;
	csz *ext = extensions; 
	csz *ext2;
	csz *varname;
	csz *varname2;
	while((varname2 = getnext_ext(&ext, name, buf))) {
		ext2 = extensions;
		while((varname = getnext_ext(&ext2, varname2, buf2))) {
			varbuf[0] = '\0';
			if((val=getenv(varname))) {
				strcat_subst(varbuf, val);
				return varbuf;
			}
		}
	}
	while((varname = getnext_ext(&ext, name, buf))) {
		varbuf[0] = '\0';
		if((val=getenv(varname))) {
			strcat_subst(varbuf, val);
			return varbuf;
		}
	}
	return get_flags(name, varbuf);
}


void
fprint_var(FILE *f, csz *varname)
{
	sz *val;
	if(!(val = getenv(varname)))
		return;
	fprintf(f, "%s = ", varname);
	while(*val) {
		if(*val == '$')
			fputs("$$", f);
		else
			fputc(*val, f);
		val++;
	}
	fputc('\n', f);
}


void
vconfig(csz *fname)
{
	csz *allext;
	csz *allvar;
	csz *ext;
	csz *var;
	sz varname[VNAME_MAX];
	FILE *fcon;
	sz varstr[VNAME_MAX], extstr[VNAME_MAX], extstr2[VNAME_MAX];

	fcon = fopen(fname, "w");
	allvar = getenv("Variables");
	while((var = getnext(&allvar, varstr))) {
		fprint_var(fcon, var);
		allext = getenv("Extensions");
		while((ext = getnext(&allext, extstr))) {
			sprintf(varname, "%s_%s", var, ext);
			fprint_var(fcon, varname);
			csz *ext2;
			csz *allext2 = getenv("Extensions");
			while((ext2 = getnext(&allext2, extstr2))) {
				sprintf(varname, "%s_%s_%s", var, ext, ext2);
				fprint_var(fcon, varname);
			}
		}
	}
	fclose(fcon);
}



void
mconfig(csz *destfname, csz *fname)
{
	DirQ *dqp;
	FILE *fmk;
	FILE *f;
	sz allpath[CMD_MAX];
	sz path[PATH_MAX];
	bool missing;
	sz buf[BUF_MAX];
	int n;

	strcpy(allpath, "");
	dqp = dirs.next();
	missing = true;
	fmk = fopen(destfname, "a");
	while(dqp) {
		sprintf(path, "%s%s/%s%s", 
			dqp->path(), dqp->name(), fname, dqp->ext());
		if((f = fopen(path, "r"))) {
			sz linkpath[PATH_MAX];
			sprintf(linkpath, "%s%s", fname, dqp->ref());
			unlink(linkpath);
			if(symlink(path, linkpath) < 0) {
				sz err[PATH_MAX*2+16];
				sprintf(err, "symlink(%s, %s)",
					path, linkpath);
				perror(err);
			}
			strcat(allpath, " ");
			fprintf(fmk, "%s: %s\n", destfname, linkpath);
			while((n = fread(buf, 1, sizeof(buf), f)) > 0)
				fwrite(buf, 1, n, fmk); 
			fclose(f);
			fprintf(fmk, "\n\n");
			strcat(allpath, " ");
			strcat(allpath, path);
			missing = false;
		}
		dqp = dqp->next();
	}
	if(missing)
		printf("Warning: missing %s\n", fname);
	fclose(fmk);
	printf("Configured %s:\n", destfname);
	fflush(stdout);
	if(allpath[0]) {
		csz *cp = allpath;
		csz *path;
		while((path = getnext(&cp, buf)))
			printf("\t%s\n", path);
	}
}



void
config_from_source(FILE *f, bool prescan)
{
	sz buf[LINE_MAX];
	csz *cp;
	sz argbuf[LINE_MAX];
	bool add;
	sz varname[VNAME_MAX];
	csz *val;

	while(fgets(buf, sizeof(buf), f)) {
		if(buf[0] != '#') 
			continue;
		cp = &buf[1];
		if(!getnext(&cp, argbuf) || strcmp(argbuf, "pragma") != 0) {
			if(prescan && strcmp(argbuf, "include") == 0) 
				return;
			else
				continue;
		}
		if(!getnext(&cp, argbuf) || strcmp(argbuf, "Make") != 0)
			continue;
		if(!getnext(&cp, varname))
			goto bad;
		if(!getnext(&cp, argbuf))
			goto bad;
		if(strcmp(argbuf, "+=") == 0)
			add = true;
		else if(strcmp(argbuf, "=") == 0)
			add = false;
		else
			goto bad;
		while(*cp <= ' ' && *cp != '\n' && *cp)
			cp++;
		val = cp;
		while(*cp && *cp != '\n')
			cp++;
		*(sz*)cp = '\0';
		if(add) {	
			argbuf[0] = '\0';
			cp = getenv(varname);
			if(cp) {
				strcpy(argbuf, cp);
				if(argbuf[0])
					strcat(argbuf, " ");
			}
			strcat(argbuf, val);
			val = argbuf;
		}
		//setenv(varname, val, true);
		{
			sz buf[65536];
			sprintf(buf, "%s=%s", varname, val);
			putenv(strdup(buf));
		}
	}
	return;
bad:
	printf("Bad pragma: %s\n", buf);
	exit(1);
}



bool
make_rules(csz *fname, csz *basename, csz *dest_ext)
// use Rules_xx_xx and Command_xx to create .mak files
{
	sz destname[NAME_MAX];
	sz makename[NAME_MAX];
	sz oldname[NAME_MAX];
	sz depname[NAME_MAX];
	sz cmd[CMD_MAX];
	sz str[CMD_MAX];
	FILE *fsrc;
	FILE *fmak;
	csz *cp;

	dest_makefile = false;
	sprintf(destname, "%s.%s", basename, dest_ext);
	sprintf(makename, "%s.%s.mak", basename, dest_ext);
	sprintf(oldname, "%s.%s.mak.old", basename, dest_ext);
	depname[0] = '\0';
	target_name = destname;
	source_name = fname;
	base_name = basename;
	
	strcat_subst(extensions, get_var("Type", str));

	cp = get_var("CppMakedepend", str);
	if(cp[0]) {
		sprintf(cmd, "%s -D__FCONFIG__ -D__MAKEDEPEND__", cp);
		fprintf(logfile, "CppMakedepend %s: %s\n", fname, cmd);
		sprintf(depname, "%s.d", basename);
	} else {
		sprintf(cmd, "%s -D__FCONFIG__", get_var("Cpp", str));
		fprintf(logfile, "Cpp %s: %s\n", fname, cmd);
	}
	fflush(stdout);
	if(!(fsrc = popen(cmd, "r"))) {
		printf("Command: %s\n", cmd);
		exit(1);
	}
	config_from_source(fsrc, false);
	pclose(fsrc);

	/* XXX - this could cause duplicate type extensions,
	 * which are minorly inefficient, but without this,
	 * type extensions that require preprocessing to reach
         * will never get used -- should probably just be
	 * checked for duplicates
         */
	strcat_subst(extensions, get_var("Type", str));

	if(depname[0]) {
		if(!(fsrc = fopen(depname, "r"))) {
			perror(depname);
			exit(1);
		}
	} else {
		sprintf(cmd, "%s -D__FCONFIG__ -D__MAKEDEPEND__",
		get_var("Makedepend", str));
		fprintf(logfile, "Makedepend %s: %s\n", fname, cmd);
		fflush(stdout);
		if(!(fsrc = popen(cmd, "r"))) {
			printf("Command: %s\n", cmd);
			exit(1);
		}
	}
	fmak = fopen(makename, "w");
	while(fgets(str, sizeof(str), fsrc)) {
		if(str[0] > ' ') {
			cp = strchr(str, ':');
			if(autoconfig)
				fprintf(fmak, "%s %s%s",
					destname, makename, cp);
			else
				fprintf(fmak, "%s%s", destname, cp);
		} else
			fputs(str, fmak);
	}
	dest_makefile = true;
	fprintf(fmak, "%s: %s\n\t%s\n", destname, 
		makename, get_var("Command", str));
	fprintf(fmak, "%s:\n\t%s \"%s\" \"%s\" %s %s.%s\n\n\n",
		makename, argvp[0], argvp[1],
		argvp[2], argvp[3], basename, dest_ext);
	fclose(fmak);
	if(depname[0]) {
		fclose(fsrc);
		unlink(depname);
	} else
		pclose(fsrc);
	sprintf(cmd, "cmp -s %s %s", makename, oldname);
	fflush(stdout);
	dest_makefile = false;
	switch(system(cmd)) {
	case 1:
	case 256:
		return true;
	case 0:
		unlink(makename);
		rename(oldname, makename);
	default:
		return false;
	} 
}



void
fconfig(csz *fname, csz *target_ext = 0)
{
	DirQ *dqp;
	FILE *fsrc;
	sz allpath[CMD_MAX];
	sz tmpname[NAME_MAX];
	sz path[PATH_MAX];
	sz altpath[PATH_MAX];
	sz namestr[NAME_MAX];
	sz origpath[PATH_MAX];
	sz origaltpath[PATH_MAX];
	sz basename[NAME_MAX];
	sz ext[NAME_MAX];
	sz rule[CMD_MAX];
	csz *src_ext;
	csz *cp;
	sz str[CMD_MAX];
	csz *allext;
	sz backspace[PATH_MAX*3];
	bool orig;
	bool alt;
	bool first;

	dest_makefile = false;

	printf("%s", fname);
	fflush(stdout);
	cp = fname;
	for(sz *bsp = backspace; *cp; cp++) {
		*bsp++ = '\b'; *bsp++ = ' '; *bsp++ = '\b'; *bsp = '\0';
	}

	if((cp = strrchr(fname, '.'))) {
		strncpy(basename, fname, cp-fname);
		basename[cp-fname] = '\0';
		strcpy(ext, cp+1);
	} else {
		strcpy(basename, fname);
		strcpy(ext, "");
	}

	allext = getenv("Extensions");
	while((src_ext = getnext(&allext, str))) {
		sprintf(rule, "Rule_%s_%s", src_ext, ext);
		sprintf(tmpname, "%s.%s", basename, src_ext);
		if((cp=getenv(rule)) && file_exists(tmpname)) { 
			printf("%s", backspace);
			fconfig(tmpname, ext);
			return;
		}
	}
	extensions = ext;
	// sprintf(tmpname, "%s.tmp", fname);
	strcpy(allpath, "");
	convert_fname(namestr);
	dqp = dirs.prev();
	orig = false;
	first = true;
	while(dqp) {
		if(ext[0]) {
			sprintf(path, "%s%s.%s",
				basename, dqp->ref(), ext);
			sprintf(altpath, "%s%s/%s%s.%s", 
				dqp->path(), dqp->name(),
				basename, dqp->ext(), ext);
		} else {
			sprintf(path, "%s%s",
				basename, dqp->ref());
			sprintf(altpath, "%s%s/%s%s", 
				dqp->path(), dqp->name(),
				basename, dqp->ext());
		}
		alt = false;
		if(!(fsrc = fopen(path, "r"))) {
			fsrc = fopen(altpath, "r");
			alt = true;
		}
		if(fsrc) {
			sz line[LINE_MAX];
			fgets(line, sizeof line, fsrc);
			if(strlen(line) > 0)
				line[strlen(line)-1] = '\0';

			// XXX - some nasty heuristics going on here to
			// determine if this is the original machine-independent
			// file or not-- needs some rethought and simplifying
			// at some point because it could get faked out
			if(first) {	
				orig = true;
				if(!alt) {
					rename(path, altpath);
					goto bail;
				}
				if(strstr(line, "autogenerated")) {
					orig = false;
					goto bail;
				}
				sz buf[PATH_MAX];
				if(readlink(altpath, buf, sizeof buf) <= 0)
					goto bail;
				if(strncmp(buf, "m-", 2) == 0
						|| strstr(buf, "-m-"))
					orig = false;
			bail:
				if(orig) {
					strcpy(origpath, path);
					strcpy(origaltpath, altpath);
				}
			} else {
				csz *p, *e;
				if(orig) 
					rename(origaltpath, origpath);
				if(alt) 
					strcpy(path, altpath);
				p = strchr(line, '%');
				if(p) 
					e = strchr(++p, '%');
				if(p && e && e-p > 0) {
					if(orig)
						symlink(origpath, origaltpath);
					sprintf(str, "mdpatch %.*s %s %s", 
						e-p, p, fname, path);
					if(system(str) != 0) {
						perror(str);
						exit(1);
					}
					sprintf(path+strlen(path), "(%.*s)", e-p, p);
				} else {
					unlink(fname);
					if(symlink(path, fname) < 0) {
						sz err[PATH_MAX*2+16];
						sprintf(err, "symlink(%s, %s)",
							path, fname);
						perror(err);
					}
					strcpy(allpath, "");
				}
				orig = false;
			}
			fclose(fsrc);
			strcat(allpath, " ");
			strcat(allpath, path);
		} 
		dqp = dqp->prev();
		first = false; 
	}

	if((fsrc = fopen(fname, "r"))) {
		config_from_source(fsrc, true);
		fclose(fsrc);
		if(orig) 
			printf("%s\t%s\n", backspace, fname);
		else {
			printf("%s\t%s:", backspace, fname);
			csz *path;
			cp = allpath;
			first = true;
			while((path = getnext(&cp, str))) {
				if(first)
					first = false;
				else
					printf("\n\t\t");
				printf(" %s", path);
			}
			printf("\n");
		}
	} else {
		printf("%s", backspace);
		if(warnings)
			printf("\twarning: missing %s\n", fname);
	}

	if(target_ext) {
		sz makefilename[NAME_MAX];
			
		sprintf(makefilename, "%s.%s.mak", basename, target_ext);	
	
		printf("%s: ", makefilename);
		fflush(stdout);

		sprintf(rule, "Rule_%s_%s", ext, target_ext);
		cp = getenv(rule);
		rule[0] = '\0';
		strcat_subst(rule, cp);
		extensions = rule;
		bool did_cnf = make_rules(fname, basename, target_ext);

		cp = makefilename;
		for(sz *bsp = backspace; *cp; cp++) {
			*bsp++ = '\b'; *bsp++ = ' '; *bsp++ = '\b'; *bsp = '\0';
		}

		printf("\b \b\b \b%s", backspace);
		if(did_cnf)
			printf("Configured %s\n", makefilename);
	}
}


int
main(int argc, csz *const *argv)
{
	csz *makefile = 0;

	if(argc < 2 || strcmp(argv[1], "-v") == 0) {
		printf("fconfig v" fconfig_VER_STR " - " fconfig_VER_DATE "\n");
		exit(1);
	}
	
	if(argc < 2) {
		fprintf(stderr, "usage: [-v] %s machtypes [-w] [-m makefile] [file...]\n",
			argv[0]);
		exit(1);
	}
	csz *ap = getenv("Autoconfig");
	if(ap) 
		autoconfig = ap[strlen(ap)-1] != '0'
		 && !(strlen(ap) >= 5 && strcmp(&ap[strlen(ap)-5], "false") == 0);
	else
		autoconfig = false;
	argvp = argv;
	machtypes = parse_list(argv[1], &nmach);
	
	if(nmach < 1) {
		fprintf(stderr, "%s: no machine-types specified!\n", argv[0]);
		exit(1);
	}
	
	// for(int i=0; i<nmach; i++)
	//	printf("mach[%d]='%s'\n", i, machtypes[i]);

	src = parse_list(".", &nsrc);

	// src = parse_list(argv[2], &nsrc);
	// spath = argv[3];
	find_dirs();
	logfile = fopen("config.log", "a");
	for(int i=2; i<argc; i++) {
		if(strcmp(argv[i], "-m") == 0)
			makefile = argv[++i];
		else if(strcmp(argv[i], "-w") == 0)
			warnings = 0;
		else if(strcmp(argv[i], "config.var") == 0)
			vconfig(argv[i]);
		else if(makefile)
			mconfig(makefile, argv[i]);
		else {
			int status;

			if(fork() == 0) {
				fconfig(argv[i]);
				exit(0);
			} else {
				wait(&status);
				if(!WIFEXITED(status) 
				   || WEXITSTATUS(status))
					exit(WEXITSTATUS(status));
			} 
		}
	}
	fclose(logfile);
}

// strdup seems to be absent from some libraries altogether, so let's define
// it here just to be safe:

extern "C" sz*
strdup(csz *s)
{
	sz *m = (sz*)malloc(strlen(s)+1);
	if(!m)
		return 0;
	return strcpy(m, s);
}


// to keep the linker from complaining:
void *_IO_stderr_;

